"""SQLite parser plugin for Android AirGuard AirTag Tracker database files.

The AirTag Tracker database file is typically stored in:
  temp/data/data/de.seemoo.at_tracking_detection.release/databases/attd_db
"""

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class AirTagEventData(events.EventData):
  """AirGuard AirTag Tracker on Android event data.

  Attributes:
    device_address (str): Address of the AirTag device.
    device_name (str): Name of the device.
    first_discovery_time (dfdatetime.DateTimeValues): date and time the device
        was first detected.
    last_seen_time (dfdatetime.DateTimeValues): date and time the device was
        last detected.
    latitude (float): Latitude of the AirTag.
    longitude (float): Longitude of the AirTag.
    rssi (int): Received Signal Strength Indicator (RSSI).
  """

  DATA_TYPE = 'android:airtag:event'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.device_address = None
    self.device_name = None
    self.first_discovery_time = None
    self.last_seen_time = None
    self.latitude = None
    self.longitude = None
    self.rssi = None


class AirTagPlugin(interface.SQLitePlugin):
  """Android AirGuard AirTag Tracker SQLite database file parser."""

  NAME = 'android_airtag'
  DATA_FORMAT = 'AirGuard AirTag Tracker on SQLite database files'

  REQUIRED_STRUCTURE = {
      'beacon': frozenset([
          'beaconId', 'receivedAt', 'rssi', 'deviceAddress', 'latitude',
          'longitude']),
      'device': frozenset([
          'deviceId', 'uniqueId', 'address', 'name', 'firstDiscovery',
          'lastSeen', 'deviceType'])}

  QUERIES = [
      ('SELECT device.address, device.name, beacon.rssi, beacon.latitude, '
       'beacon.longitude, device.firstDiscovery, device.lastSeen FROM device '
       'INNER JOIN beacon ON device.address = beacon.deviceAddress',
       'ParseAirTagRow')]

  SCHEMAS = [{
      'android_metadata': (
          'CREATE TABLE android_metadata (locale TEXT)'),
      'beacon': (
          'CREATE TABLE beacon '
          '(beaconId INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
          'receivedAt TEXT NOT NULL, rssi INTEGER NOT NULL, '
          'deviceAddress TEXT NOT NULL, '
          'longitude REAL, latitude REAL, mfg BLOB, serviceUUIDs TEXT)'),
      'device': (
          'CREATE TABLE device '
          '(deviceId INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
          'uniqueId TEXT, address TEXT NOT NULL, '
          'name TEXT, ignore INTEGER NOT NULL, '
          'connectable INTEGER DEFAULT 0, payloadData INTEGER, '
          'firstDiscovery TEXT NOT NULL, '
          'lastSeen TEXT NOT NULL, notificationSent INTEGER NOT NULL, '
          'lastNotificationSent TEXT, '
          'deviceType TEXT)'),
      'feedback': (
          'CREATE TABLE feedback '
          '(feedbackId INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
          'notificationId INTEGER NOT NULL, location TEXT)'),
      'notification': (
          'CREATE TABLE notification '
          '(notificationId INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
          'deviceAddress TEXT NOT NULL, '
          'falseAlarm INTEGER NOT NULL, dismissed INTEGER, '
          'clicked INTEGER, createdAt TEXT NOT NULL)'),
      'room_master_table': (
          'CREATE TABLE room_master_table '
          '(id INTEGER PRIMARY KEY, identity_hash TEXT)'),
      'scan': (
          'CREATE TABLE scan '
          '(scanId INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, '
          'endDate TEXT, noDevicesFound INTEGER, duration INTEGER, '
          'isManual INTEGER NOT NULL, '
          'scanMode INTEGER NOT NULL, startDate TEXT)')}]

  def _GetDateTimeRowValue(self, parser_mediator, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.TimeElementsInMilliseconds: date and time value or None if
          not available.
    """
    iso8601_string = self._GetRowValue(query_hash, row, value_name)
    if iso8601_string is None:
      return None

    try:
      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds()
      date_time.CopyFromStringISO8601(iso8601_string)
    except ValueError as exception:
      parser_mediator.ProduceExtractionWarning(
          f'Unable to parse ${value_name:} date and time string: '
          f'{iso8601_string:s} with error: {exception!s}')
      date_time = None

    return date_time

  def ParseAirTagRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses an AirTag row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = AirTagEventData()
    event_data.device_address = self._GetRowValue(query_hash, row, 'address')
    event_data.device_name = self._GetRowValue(query_hash, row, 'name')
    event_data.first_discovery_time = self._GetDateTimeRowValue(
        parser_mediator, query_hash, row, 'firstDiscovery')
    event_data.last_seen_time = self._GetDateTimeRowValue(
        parser_mediator, query_hash, row, 'lastSeen')
    event_data.latitude = self._GetRowValue(query_hash, row, 'latitude')
    event_data.longitude = self._GetRowValue(query_hash, row, 'longitude')
    event_data.rssi = self._GetRowValue(query_hash, row, 'rssi')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(AirTagPlugin)



"""Imports for the text parser."""

from plaso.parsers.text_plugins import android_logcat
from plaso.parsers.text_plugins import apache_access
from plaso.parsers.text_plugins import apt_history
from plaso.parsers.text_plugins import atlassian_bitbucket
from plaso.parsers.text_plugins import atlassian_confluence
from plaso.parsers.text_plugins import atlassian_jira
from plaso.parsers.text_plugins import aws_elb_access
from plaso.parsers.text_plugins import bash_history
from plaso.parsers.text_plugins import bitbucket_access
from plaso.parsers.text_plugins import bitbucket_audit
from plaso.parsers.text_plugins import confluence_access
from plaso.parsers.text_plugins import cri
from plaso.parsers.text_plugins import dpkg
from plaso.parsers.text_plugins import gdrive_synclog
from plaso.parsers.text_plugins import google_logging
from plaso.parsers.text_plugins import iis
from plaso.parsers.text_plugins import ios_lockdownd
from plaso.parsers.text_plugins import ios_logd
from plaso.parsers.text_plugins import ios_sysdiag_log
from plaso.parsers.text_plugins import jira_access
from plaso.parsers.text_plugins import macos_appfirewall
from plaso.parsers.text_plugins import macos_launchd
from plaso.parsers.text_plugins import macos_securityd
from plaso.parsers.text_plugins import macos_wifi
from plaso.parsers.text_plugins import popcontest
from plaso.parsers.text_plugins import postgresql
from plaso.parsers.text_plugins import powershell_transcript
from plaso.parsers.text_plugins import santa
from plaso.parsers.text_plugins import sccm
from plaso.parsers.text_plugins import selinux
from plaso.parsers.text_plugins import setupapi
from plaso.parsers.text_plugins import skydrivelog
from plaso.parsers.text_plugins import snort_fastlog
from plaso.parsers.text_plugins import sophos_av
from plaso.parsers.text_plugins import syslog
from plaso.parsers.text_plugins import teamviewer
from plaso.parsers.text_plugins import viminfo
from plaso.parsers.text_plugins import vsftpd
from plaso.parsers.text_plugins import winfirewall
from plaso.parsers.text_plugins import xchatlog
from plaso.parsers.text_plugins import xchatscrollback
from plaso.parsers.text_plugins import zsh_extended_history



"""Shared functionality for OpenSearch output modules."""

import logging
import os

from acstore.containers import interface as containers_interface

from dfdatetime import interface as dfdatetime_interface
from dfdatetime import posix_time as dfdatetime_posix_time

from dfvfs.serializer.json_serializer import JsonPathSpecSerializer

try:
  import opensearchpy
except ImportError:
  opensearchpy = None

from plaso.lib import errors
from plaso.output import formatting_helper
from plaso.output import interface
from plaso.output import logger


# Configure the OpenSearch logger.
if opensearchpy:
  opensearch_logger = logging.getLogger('opensearchpy.trace')
  opensearch_logger.setLevel(logging.WARNING)


class SharedOpenSearchFieldFormattingHelper(
    formatting_helper.FieldFormattingHelper):
  """Shared OpenSearch output module field formatting helper."""

  # Maps the name of a fields to a a callback function that formats
  # the field value.
  _FIELD_FORMAT_CALLBACKS = {
      'datetime': '_FormatDateTime',
      'display_name': '_FormatDisplayName',
      'inode': '_FormatInode',
      'message': '_FormatMessage',
      'source_long': '_FormatSource',
      'source_short': '_FormatSourceShort',
      'tag': '_FormatTag',
      'timestamp': '_FormatTimestamp',
      'timestamp_desc': '_FormatTimestampDescription',
      'yara_match': '_FormatYaraMatch'}

  # The field format callback methods require specific arguments hence
  # the check for unused arguments is disabled here.
  # pylint: disable=unused-argument

  def _FormatDateTime(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a date and time field in ISO 8601 format.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: date and time field.
    """
    date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
        timestamp=event.timestamp)
    return date_time.CopyToDateTimeStringISO8601()

  def _FormatTag(self, output_mediator, event_tag):
    """Formats an event tag field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event_tag (EventTag): event tag or None if not set.

    Returns:
      list[str]: event tag labels.
    """
    return getattr(event_tag, 'labels', None) or []

  def _FormatInode(self, output_mediator, event, event_data, event_data_stream):
    """Formats an inode field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: inode field.
    """
    inode = getattr(event_data, 'inode', None)
    if isinstance(inode, int):
      inode = f'{inode:d}'

    return inode

  def _FormatTimestamp(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a timestamp field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      int: timestamp field.
    """
    return event.timestamp

  def _FormatTimestampDescription(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a timestamp description field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: timestamp description field.
    """
    return event.timestamp_desc

  def _FormatYaraMatch(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a Yara match field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      list[str]: Yara match field.
    """
    return getattr(event_data_stream, 'yara_match', None) or []

  # pylint: enable=unused-argument

  def GetFormattedField(
      self, output_mediator, field_name, event, event_data, event_data_stream,
      event_tag):
    """Formats the specified field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_name (str): name of the field.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      object: value of the field or None if not set.
    """
    callback_name = self._FIELD_FORMAT_CALLBACKS.get(field_name)
    if callback_name == '_FormatTag':
      return self._FormatTag(output_mediator, event_tag)

    callback_function = None
    if callback_name:
      callback_function = getattr(self, callback_name, None)

    if callback_function:
      output_value = callback_function(
          output_mediator, event, event_data, event_data_stream)
    elif hasattr(event_data_stream, field_name):
      output_value = getattr(event_data_stream, field_name, None)
    else:
      output_value = getattr(event_data, field_name, None)

    return output_value


class SharedOpenSearchOutputModule(interface.OutputModule):
  """Shared functionality for an OpenSearch output module."""

  # pylint: disable=abstract-method

  NAME = 'opensearch_shared'

  SUPPORTS_ADDITIONAL_FIELDS = True
  SUPPORTS_CUSTOM_FIELDS = True

  _DEFAULT_FLUSH_INTERVAL = 1000

  # Number of seconds to wait before a request to OpenSearch is timed out.
  _DEFAULT_REQUEST_TIMEOUT = 300

  _DEFAULT_FIELD_NAMES = [
      'datetime',
      'display_name',
      'message',
      'source_long',
      'source_short',
      'tag',
      'timestamp',
      'timestamp_desc']

  def __init__(self):
    """Initializes an output module."""
    super().__init__()
    self._client = None
    self._custom_fields = {}
    self._event_documents = []
    self._field_names = self._DEFAULT_FIELD_NAMES
    self._field_formatting_helper = SharedOpenSearchFieldFormattingHelper()
    self._flush_interval = self._DEFAULT_FLUSH_INTERVAL
    self._host = None
    self._index_name = None
    self._mappings = None
    self._number_of_buffered_events = 0
    self._password = None
    self._port = None
    self._username = None
    self._use_ssl = None
    self._ca_certs = None
    self._url_prefix = None

  def _Connect(self):
    """Connects to an OpenSearch server.

    Raises:
      RuntimeError: if the OpenSearch version is not supported or the server
          cannot be reached.
    """
    opensearch_host = {'host': self._host, 'port': self._port}

    if self._url_prefix:
      opensearch_host['url_prefix'] = self._url_prefix

    opensearch_http_auth = None
    if self._username is not None:
      opensearch_http_auth = (self._username, self._password)

    self._client = opensearchpy.OpenSearch(
        [opensearch_host],
        http_auth=opensearch_http_auth,
        use_ssl=self._use_ssl,
        ca_certs=self._ca_certs)

    logger.debug((
        f'Connected to OpenSearch server: {self._host:s} port: {self._port:d} '
        f'URL prefix: {self._url_prefix!s}.'))

  def _CreateIndexIfNotExists(self, index_name, mappings):
    """Creates an OpenSearch index if it does not exist.

    Args:
      index_name (str): mame of the index.
      mappings (dict[str, object]): mappings of the index.

    Raises:
      RuntimeError: if the OpenSearch index cannot be created.
    """
    try:
      # Try opensearch-py >= 2.5.0 calling convention.
      index_exists = self._client.indices.exists(index=index_name)
    except TypeError:
      # pylint: disable=missing-kwoa,too-many-function-args
      index_exists = self._client.indices.exists(index_name)

    try:
      if not index_exists:
        self._client.indices.create(
            body={'mappings': mappings}, index=index_name)
    except opensearchpy.exceptions.ConnectionError as exception:
      raise RuntimeError(
          f'Unable to create OpenSearch index with error: {exception!s}')

  def _FlushEvents(self):
    """Inserts the buffered event documents into OpenSearch."""
    try:
      # pylint: disable=unexpected-keyword-arg
      bulk_arguments = {
          'body': self._event_documents,
          'index': self._index_name,
          'request_timeout': self._DEFAULT_REQUEST_TIMEOUT}

      self._client.bulk(**bulk_arguments)

    except (ValueError,
            opensearchpy.exceptions.OpenSearchException) as exception:
      # Ignore problematic events
      logger.warning(f'Unable to bulk insert with error: {exception!s}')

    logger.debug(
        'Inserted {self._number_of_buffered_events:d} events into OpenSearch')

    self._event_documents = []
    self._number_of_buffered_events = 0

  def _SanitizeField(self, data_type, attribute_name, field):
    """Sanitizes a field for output.

    Args:
      data_type (str): event data type.
      attribute_name (str): name of the event attribute.
      field (object): value of the field to sanitize.

    Returns:
      object: sanitized value of the field.
    """
    # Some parsers have written bytes values to storage.
    if isinstance(field, bytes):
      field = field.decode('utf-8', 'replace')
      logger.warning((
          f'Found bytes value for attribute: {attribute_name:s} of data type: '
          f'{data_type!s}. Value was converted to UTF-8: "{field:s}"'))

    return field

  def Close(self):
    """Closes connection to OpenSearch.

    Inserts any remaining buffered event documents.
    """
    self._FlushEvents()

    self._client = None

  def GetFieldValues(
      self, output_mediator, event, event_data, event_data_stream, event_tag):
    """Retrieves the output field values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      dict[str, str]: output field values per name.
    """
    event_values = {}

    if event_data:
      for attribute_name, attribute_value in event_data.GetAttributes():
        # Ignore attribute container identifier and date and time values.
        if isinstance(attribute_value, (
            containers_interface.AttributeContainerIdentifier,
            dfdatetime_interface.DateTimeValues)):
          continue

        if (isinstance(attribute_value, list) and attribute_value and
            isinstance(attribute_value[0],
                       dfdatetime_interface.DateTimeValues)):
          continue

        # Ignore protected internal only attributes.
        if attribute_name[0] == '_' and attribute_name != '_parser_chain':
          continue

        # Output _parser_chain as parser for backwards compatibility.
        if attribute_name == '_parser_chain':
          attribute_name = 'parser'

        event_values[attribute_name] = attribute_value

    if event_data_stream:
      for attribute_name, attribute_value in event_data_stream.GetAttributes():
        event_values[attribute_name] = attribute_value

    for attribute_name in self._field_names:
      if attribute_name not in event_values:
        event_values[attribute_name] = None

    field_values = {}
    for attribute_name, attribute_value in event_values.items():
      if attribute_name == 'path_spec':
        try:
          field_value = JsonPathSpecSerializer.WriteSerialized(attribute_value)
        except TypeError:
          continue

      else:
        field_value = self._field_formatting_helper.GetFormattedField(
            output_mediator, attribute_name, event, event_data,
            event_data_stream, event_tag)

      if field_value is None and attribute_name in self._custom_fields:
        field_value = self._custom_fields.get(attribute_name)

      if field_value is None:
        field_value = '-'

      field_values[attribute_name] = self._SanitizeField(
          event_data.data_type, attribute_name, field_value)

    return field_values

  def SetAdditionalFields(self, field_names):
    """Sets the names of additional fields to output.

    Args:
      field_names (list[str]): names of additional fields to output.
    """
    self._field_names.extend(field_names)

  def SetCustomFields(self, field_names_and_values):
    """Sets the names and values of custom fields to output.

    Args:
      field_names_and_values (list[tuple[str, str]]): names and values of
          custom fields to output.
    """
    self._custom_fields = dict(field_names_and_values)
    self._field_names.extend(self._custom_fields.keys())

  def SetFlushInterval(self, flush_interval):
    """Sets the flush interval.

    Args:
      flush_interval (int): number of events to buffer before doing a bulk
          insert.
    """
    self._flush_interval = flush_interval
    logger.debug(f'OpenSearch flush interval: {flush_interval:d}')

  def SetIndexName(self, index_name):
    """Sets the index name.

    Args:
      index_name (str): name of the index.
    """
    self._index_name = index_name
    logger.debug(f'OpenSearch index name: {index_name:s}')

  def SetMappings(self, mappings):
    """Sets the mappings.

    Args:
      mappings (dict[str, object]): mappings of the index.
    """
    self._mappings = mappings

  def SetPassword(self, password):
    """Sets the password.

    Args:
      password (str): password to authenticate with.
    """
    self._password = password
    logger.debug('OpenSearch password: ********')

  def SetServerInformation(self, server, port):
    """Sets the server information.

    Args:
      server (str): IP address or hostname of the server.
      port (int): Port number of the server.
    """
    self._host = server
    self._port = port
    logger.debug(f'OpenSearch server: {server!s} port: {port:d}')

  def SetUsername(self, username):
    """Sets the username.

    Args:
      username (str): username to authenticate with.
    """
    self._username = username
    logger.debug(f'OpenSearch username: {username!s}')

  def SetUseSSL(self, use_ssl):
    """Sets the use of ssl.

    Args:
      use_ssl (bool): enforces use of ssl.
    """
    self._use_ssl = use_ssl
    logger.debug(f'OpenSearch use SSL/TLS: {use_ssl!s}')

  def SetCACertificatesPath(self, ca_certificates_path):
    """Sets the path to the CA certificates.

    Args:
      ca_certificates_path (str): path to file containing a list of root
        certificates to trust.

    Raises:
      BadConfigOption: if the CA certificates file does not exist.
    """
    if not ca_certificates_path:
      return

    if not os.path.exists(ca_certificates_path):
      raise errors.BadConfigOption(
          f'No such certificate file: {ca_certificates_path:s}')

    self._ca_certs = ca_certificates_path
    logger.debug(f'OpenSearch certificate file: {ca_certificates_path:s}')

  def SetURLPrefix(self, url_prefix):
    """Sets the URL prefix.

    Args:
      url_prefix (str): URL prefix.
    """
    self._url_prefix = url_prefix
    logger.debug('OpenSearch URL prefix: {0!s}')



"""The timeliner, which is used to generate events from event data."""

import collections
import copy
import datetime
import decimal
import os
import pytz

from dfdatetime import interface as dfdatetime_interface
from dfdatetime import semantic_time as dfdatetime_semantic_time

from plaso.containers import events
from plaso.containers import warnings
from plaso.engine import yaml_timeliner_file
from plaso.lib import definitions


class EventDataTimeliner:
  """The event data timeliner.

  Attributes:
    number_of_produced_events (int): number of produced events.
    parsers_counter (collections.Counter): number of events per parser or
        parser plugin.
  """

  _DEFAULT_TIME_ZONE = pytz.UTC

  _INT64_MIN = -1 << 63
  _INT64_MAX = (1 << 63) - 1

  _TIMELINER_CONFIGURATION_FILENAME = 'timeliner.yaml'

  def __init__(
      self, data_location=None, preferred_year=None,
      system_configurations=None):
    """Initializes an event data timeliner.

    Args:
      data_location (Optional[str]): path of the timeliner configuration file.
      preferred_year (Optional[int]): preferred initial year value for date-less
          date and time values.
      system_configurations (Optional[list[SystemConfigurationArtifact]]):
          system configurations.
    """
    super().__init__()
    self._attribute_mappings = {}
    self._base_dates = {}
    self._current_date = self._GetCurrentDate()
    self._data_location = data_location
    self._place_holder_event = set()
    self._preferred_time_zone = None
    self._preferred_year = preferred_year
    self._time_zone_per_path_spec = None

    self.number_of_produced_events = 0
    self.parsers_counter = collections.Counter()

    self._CreateTimeZonePerPathSpec(system_configurations)
    self._ReadConfigurationFile()

  def _CreateTimeZonePerPathSpec(self, system_configurations):
    """Creates the time zone per path specification lookup table.

    Args:
      system_configurations (list[SystemConfigurationArtifact]): system
          configurations.
    """
    self._time_zone_per_path_spec = {}
    for system_configuration in system_configurations or []:
      if system_configuration.time_zone:
        for path_spec in system_configuration.path_specs:
          if path_spec.parent:
            self._time_zone_per_path_spec[path_spec.parent] = (
                system_configuration.time_zone)

  def _GetBaseDate(self, storage_writer, event_data):
    """Retrieves the base date.

    Args:
      storage_writer (StorageWriter): storage writer.
      event_data (EventData): event data.

    Returns:
      tuple[int, int, int]: base date, as a tuple of year, month, day of month.
    """
    # If preferred year is set considered it a user override, otherwise try
    # to determine the year based on the date-less log helper or fallback to
    # the current year.

    if self._preferred_year:
      return self._preferred_year, 0, 0

    current_date = self._current_date

    event_data_stream_identifier = event_data.GetEventDataStreamIdentifier()
    if not event_data_stream_identifier:
      return current_date[0], 0, 0

    lookup_key = event_data_stream_identifier.CopyToString()

    base_date = self._base_dates.get(lookup_key)
    if base_date:
      return base_date

    filter_expression = f'_event_data_stream_identifier == "{lookup_key:s}"'
    date_less_log_helpers = list(storage_writer.GetAttributeContainers(
        events.DateLessLogHelper.CONTAINER_TYPE,
        filter_expression=filter_expression))
    if not date_less_log_helpers:
      self._ProduceTimeliningWarning(storage_writer, event_data, (
          f'missing date-less log helper, defaulting to date: '
          f'{current_date[0]:d}-{current_date[1]:d}-{current_date[2]:d}'))

      base_date = (current_date[0], 0, 0)

    else:
      date_less_log_helper = date_less_log_helpers[0]

      earliest_date = date_less_log_helper.GetEarliestDate()
      last_relative_date = date_less_log_helper.GetLastRelativeDate()
      latest_date = date_less_log_helper.GetLatestDate()

      if date_less_log_helper.granularity == (
          date_less_log_helper.GRANULARITY_NO_YEAR):
        current_date = (current_date[0], 0, 0)

      if earliest_date is None or last_relative_date is None:
        last_date = None
      else:
        last_date = tuple(map(
            lambda earliest, last_relative: earliest + last_relative,
            earliest_date, last_relative_date))

      if earliest_date is None and latest_date is None:
        self._ProduceTimeliningWarning(storage_writer, event_data, (
            f'missing earliest and latest date in date-less log helper, '
            f'defaulting to date: {current_date[0]:d}-{current_date[1]:d}-'
            f'{current_date[2]:d}'))

        base_date = current_date

      elif last_date < current_date:
        base_date = earliest_date

      elif latest_date < current_date:
        self._ProduceTimeliningWarning(storage_writer, event_data, (
            f'earliest date: {earliest_date[0]:d}-{earliest_date[1]:d}-'
            f'{earliest_date[2]:d} as base date would exceed : '
            f'{current_date[0]:d}-{current_date[1]:d}-{current_date[2]:d} + '
            f'{last_relative_date[0]:d}-{last_relative_date[1]:d}-'
            f'{last_relative_date[2]:d}, using latest date: {latest_date[0]:d}-'
            f'{latest_date[1]:d}-{latest_date[2]:d}'))

        base_date = tuple(map(
            lambda latest, last_relative: latest - last_relative,
            latest_date, last_relative_date))

      else:
        self._ProduceTimeliningWarning(storage_writer, event_data, (
            f'earliest date: {earliest_date[0]:d}-{earliest_date[1]:d}-'
            f'{earliest_date[2]:d} and latest date: {latest_date[0]:d}-'
            f'{latest_date[1]:d}-{latest_date[2]:d} as base date would exceed '
            f'date: {current_date[0]:d}-{current_date[1]:d}-'
            f'{current_date[2]:d} + {last_relative_date[0]:d}-'
            f'{last_relative_date[1]:d}-{last_relative_date[2]:d}, using date: '
            f'{current_date[0]:d}-{current_date[1]:d}-{current_date[2]:d}'))

        base_date = tuple(map(
            lambda current, last_relative: current - last_relative,
            current_date, last_relative_date))

    self._base_dates[lookup_key] = base_date

    return base_date

  def _GetCurrentDate(self):
    """Retrieves current date.

    Returns:
      tuple[int, int, int]: current date, as a tuple of year, month, day of
          month.
    """
    datetime_object = datetime.datetime.now(pytz.UTC)
    return datetime_object.year, datetime_object.month, datetime_object.day

  def _GetEvent(
      self, storage_writer, event_data, event_data_stream, date_time,
      date_time_description):
    """Retrieves an event.

    Args:
      storage_writer (StorageWriter): storage writer.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      date_time (dfdatetime.DateTimeValues): date and time values.
      date_time_description (str): description of the meaning of the date and
          time values.

    Returns:
      EventObject: event.
    """
    timestamp = None
    if date_time.is_delta:
      base_date = self._GetBaseDate(storage_writer, event_data)

      try:
        date_time = date_time.NewFromDeltaAndDate(*base_date)
      except ValueError as exception:
        self._ProduceTimeliningWarning(
            storage_writer, event_data, str(exception))

        date_time = dfdatetime_semantic_time.InvalidTime()
        timestamp = 0

    if timestamp is None:
      try:
        timestamp = date_time.GetPlasoTimestamp()
      except decimal.InvalidOperation as exception:
        self._ProduceTimeliningWarning(
            storage_writer, event_data, str(exception))

        date_time = dfdatetime_semantic_time.InvalidTime()
        timestamp = 0

    if timestamp is None:
      self._ProduceTimeliningWarning(
          storage_writer, event_data, 'unable to determine timestamp')

      date_time = dfdatetime_semantic_time.InvalidTime()
      timestamp = 0

    # Check for out of bounds timestamps, for example if a data format has
    # changed and the conversion leads to an incorrect large integer value.
    # Integer values that are larger than 64-bit will cause an OverflowError
    # in the SQLite storage.

    elif timestamp < self._INT64_MIN or timestamp > self._INT64_MAX:
      self._ProduceTimeliningWarning(
          storage_writer, event_data, 'timestamp out of bounds')

      date_time = dfdatetime_semantic_time.InvalidTime()
      timestamp = 0

    else:
      if date_time.is_local_time:
        time_zone = None
        if date_time.time_zone_hint:
          # TODO: cache time zones per hint.
          try:
            time_zone = pytz.timezone(date_time.time_zone_hint)
          except pytz.UnknownTimeZoneError:
            self._ProduceTimeliningWarning(storage_writer, event_data, (
                f'unsupported time zone hint: {date_time.time_zone_hint:s}, '
                f'using default time zone'))

        if not time_zone and event_data_stream:
          try:
            time_zone = self._GetTimeZoneByPathSpec(event_data_stream.path_spec)
          except pytz.UnknownTimeZoneError:
            self._ProduceTimeliningWarning(storage_writer, event_data, (
                f'unsupported system time zone: {date_time.time_zone_hint:s}, '
                f'using default time zone'))

        if not time_zone:
          time_zone = self._preferred_time_zone or self._DEFAULT_TIME_ZONE

        date_time = copy.deepcopy(date_time)
        date_time.is_local_time = False

        if time_zone != pytz.UTC:
          datetime_object = datetime.datetime(
              1970, 1, 1, 0, 0, 0, 0, tzinfo=None)
          datetime_object += datetime.timedelta(microseconds=timestamp)

          datetime_delta = time_zone.utcoffset(datetime_object, is_dst=False)
          seconds_delta = int(datetime_delta.total_seconds())
          timestamp -= seconds_delta * definitions.MICROSECONDS_PER_SECOND

          date_time.time_zone_offset = seconds_delta // 60

    event = events.EventObject()
    event.date_time = date_time
    event.timestamp = timestamp
    event.timestamp_desc = date_time_description

    event_data_identifier = event_data.GetIdentifier()
    event.SetEventDataIdentifier(event_data_identifier)

    return event

  def _GetTimeZoneByPathSpec(self, path_spec):
    """Retrieves a time zone for a specific path specification.

    Args:
      path_spec (dfvfs.PathSpec): path specification.

    Returns:
      pytz.tzfile: time zone or None if not available.

    Raises:
      pytz.UnknownTimeZoneError: if the time zone is unknown.
    """
    if not path_spec or not path_spec.parent:
      return None

    time_zone = self._time_zone_per_path_spec.get(path_spec.parent)
    if not time_zone:
      return None

    if isinstance(time_zone, str):
      try:
        time_zone = pytz.timezone(time_zone)
        self._time_zone_per_path_spec[path_spec.parent] = time_zone
      except pytz.UnknownTimeZoneError as exeception:
        self._time_zone_per_path_spec[path_spec.parent] = None
        raise exeception

    return time_zone

  def _ProduceTimeliningWarning(self, storage_writer, event_data, message):
    """Produces a timelining warning.

    Args:
      storage_writer (StorageWriter): storage writer.
      event_data (EventData): event data.
      message (str): message of the warning.
    """
    parser_chain = getattr(event_data, '_parser_chain', None)
    path_spec = None

    event_data_stream_identifier = event_data.GetEventDataStreamIdentifier()
    if event_data_stream_identifier:
      event_data_stream = storage_writer.GetAttributeContainerByIdentifier(
          events.EventDataStream.CONTAINER_TYPE, event_data_stream_identifier)
      if event_data_stream:
        path_spec = event_data_stream.path_spec

    warning = warnings.TimeliningWarning(
        message=message, parser_chain=parser_chain, path_spec=path_spec)
    storage_writer.AddAttributeContainer(warning)

  def _ReadConfigurationFile(self):
    """Reads a timeliner configuration file.

    Raises:
      KeyError: if the attribute mappings are already set for the corresponding
          data type.
    """
    path = os.path.join(
        self._data_location, self._TIMELINER_CONFIGURATION_FILENAME)

    configuration_file = yaml_timeliner_file.YAMLTimelinerConfigurationFile()
    for timeliner_definition in configuration_file.ReadFromFile(path):
      if timeliner_definition.data_type in self._attribute_mappings:
        raise KeyError((
            f'Attribute mappings for data type: '
            f'{timeliner_definition.data_type:s} already set.'))

      self._attribute_mappings[timeliner_definition.data_type] = (
          timeliner_definition.attribute_mappings)

      if timeliner_definition.place_holder_event:
        self._place_holder_event.add(timeliner_definition.data_type)

  def ProcessEventData(self, storage_writer, event_data, event_data_stream):
    """Generate events from event data.

    Args:
      storage_writer (StorageWriter): storage writer.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
    """
    self.number_of_produced_events = 0

    attribute_mappings = self._attribute_mappings.get(
        event_data.data_type) or {}
    if (not attribute_mappings and
        event_data.data_type not in self._place_holder_event):
      return

    parser_name = None
    parser_chain = getattr(event_data, '_parser_chain', None)
    if parser_chain:
      parser_name = parser_chain.rsplit('/', maxsplit=1)[-1]

    number_of_events = 0
    for attribute_name, time_description in attribute_mappings.items():
      attribute_values = getattr(event_data, attribute_name, None) or []
      if not isinstance(attribute_values, list):
        attribute_values = [attribute_values]

      for attribute_value in attribute_values:
        if not isinstance(attribute_value, dfdatetime_interface.DateTimeValues):
          message = f'unsupported date time attribute: {attribute_name:s}'
          self._ProduceTimeliningWarning(storage_writer, event_data, message)
          continue

        event = self._GetEvent(
            storage_writer, event_data, event_data_stream, attribute_value,
            time_description)

        try:
          storage_writer.AddAttributeContainer(event)
        except OverflowError as exception:
          message = f'unable to add event with error: {exception!s}'
          self._ProduceTimeliningWarning(storage_writer, event_data, message)
          continue

        number_of_events += 1

        if parser_name:
          self.parsers_counter[parser_name] += 1
        self.parsers_counter['total'] += 1

        self.number_of_produced_events += 1

    # Create a place holder event for event_data without date and time
    # values to map.
    if (not number_of_events and
        event_data.data_type in self._place_holder_event):
      date_time = dfdatetime_semantic_time.NotSet()
      event = self._GetEvent(
          storage_writer, event_data, event_data_stream, date_time,
          definitions.TIME_DESCRIPTION_NOT_A_TIME)

      storage_writer.AddAttributeContainer(event)

      if parser_name:
        self.parsers_counter[parser_name] += 1
      self.parsers_counter['total'] += 1

      self.number_of_produced_events += 1

  def SetPreferredTimeZone(self, time_zone_string):
    """Sets the preferred time zone for zone-less date and time values.

    Args:
      time_zone_string (str): time zone such as "Europe/Amsterdam" or None if
          the time zone determined by preprocessing or the default should be
          used.

    Raises:
      ValueError: if the time zone is not supported.
    """
    time_zone = None
    if time_zone_string:
      try:
        time_zone = pytz.timezone(time_zone_string)
      except pytz.UnknownTimeZoneError:
        raise ValueError(f'Unsupported time zone: {time_zone_string!s}')

    self._preferred_time_zone = time_zone



"""Windows Prefetch custom event formatter helpers."""

from plaso.formatters import interface
from plaso.formatters import manager


class WindowsPrefetchPathHintsFormatterHelper(
    interface.CustomEventFormatterHelper):
  """Windows Prefetch path hints formatter helper."""

  IDENTIFIER = 'windows_prefetch_path_hints'

  def FormatEventValues(self, output_mediator, event_values):
    """Formats event values using the helper.

    Args:
      output_mediator (OutputMediator): output mediator.
      event_values (dict[str, object]): event values.
    """
    path_hints = event_values.get('path_hints')
    if path_hints:
      event_values['path_hints'] = '; '.join(path_hints)


class WindowsPrefetchVolumesStringFormatterHelper(
    interface.CustomEventFormatterHelper):
  """Windows Prefetch volumes string formatter helper."""

  IDENTIFIER = 'windows_prefetch_volumes_string'

  def FormatEventValues(self, output_mediator, event_values):
    """Formats event values using the helper.

    Args:
      output_mediator (OutputMediator): output mediator.
      event_values (dict[str, object]): event values.
    """
    number_of_volumes = event_values.get('number_of_volumes', 0)
    volume_serial_numbers = event_values.get('volume_serial_numbers')
    volume_device_paths = event_values.get('volume_device_paths')

    volumes_strings = []
    for volume_index in range(0, number_of_volumes):
      if not volume_serial_numbers:
        volume_serial_number = 'UNKNOWN'
      else:
        volume_serial_number = f'0x{volume_serial_numbers[volume_index]:08X}'

      if not volume_device_paths:
        volume_device_path = 'UNKNOWN'
      else:
        volume_device_path = volume_device_paths[volume_index]

      volumes_strings.append((
          f'volume: {volume_index + 1:d} [serial number: '
          f'{volume_serial_number:s}, device path: {volume_device_path:s}]'))

    if volumes_strings:
      event_values['volumes_string'] = ', '.join(volumes_strings)


manager.FormattersManager.RegisterEventFormatterHelpers([
    WindowsPrefetchPathHintsFormatterHelper,
    WindowsPrefetchVolumesStringFormatterHelper])



"""SQLite parser plugin for Android application launch database files."""

from dfdatetime import java_time as dfdatetime_java_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class AndroidAppLaunch(events.EventData):
  """Android application launch event data.

  Attributes:
    identifier (int): identifier.
    launch_location_identifier (int): identifier of the location where the app
        was launched.
    package_name (str): name of the package of the app.
    prediction_rank (int): relevance of the prediction.
    prediction_source_identifier (int): identifier of the source of prediction.
    prediction_ui_surface_identifier (int): identifier of the UI surface where
        prediction was made.
    start_time (dfdatetime.DateTimeValues): date and time the application was
        started (or launched).
  """

  DATA_TYPE = 'android:sqlite:app_launch'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.identifier = None
    self.launch_location_identifier = None
    self.package_name = None
    self.prediction_rank = None
    self.prediction_source_identifier = None
    self.prediction_ui_surface_identifier = None
    self.start_time = None


class AndroidAppLaunchPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Android application launch database files."""

  NAME = 'android_app_launch'
  DATA_FORMAT = (
      'Android application launch SQLite database (SimpleStorage) file')

  REQUIRED_STRUCTURE = {
      'EchoAppLaunchMetricsEvents': frozenset([
          'timestampMillis', 'packageName', 'launchLocationId',
          'predictionUiSurfaceId', 'predictionSourceId', 'predictionRank',
          'id'])}

  QUERIES = [(
      ('SELECT timestampMillis, packageName, launchLocationId, '
       'predictionUiSurfaceId, predictionSourceId, predictionRank, id '
       'FROM EchoAppLaunchMetricsEvents'), 'ParseAppLaunchRow')]

  SCHEMAS = {
      'EchoAppLaunchMetricsEvents': (
          'CREATE TABLE `EchoAppLaunchMetricsEvents` '
          '(`timestampMillis` INTEGER NOT NULL, `packageName` TEXT NOT NULL, '
          '`launchLocationId` INTEGER NOT NULL, `predictionUiSurfaceId` '
          'INTEGER NOT NULL, `predictionSourceId` INTEGER NOT NULL, '
          '`predictionRank` INTEGER NOT NULL, `id` INTEGER PRIMARY '
          'KEY AUTOINCREMENT NOT NULL, FOREIGN KEY(`packageName`) REFERENCES '
          '`Packages`(`packageName`) ON UPDATE CASCADE ON DELETE CASCADE )'),
      'Packages': (
          'CREATE TABLE `Packages` (`packageName` TEXT NOT NULL, '
          '`loggablePackageName` TEXT NOT NULL, PRIMARY KEY(`packageName`))')}

  def _GetTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.

    Returns:
      dfdatetime.JavaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_java_time.JavaTime(timestamp=timestamp)

  # pylint: disable=unused-argument
  def ParseAppLaunchRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses an event record row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = AndroidAppLaunch()
    event_data.identifier = self._GetRowValue(query_hash, row, 'id')
    event_data.launch_location_identifier = self._GetRowValue(
        query_hash, row, 'launchLocationId')
    event_data.package_name = self._GetRowValue(query_hash, row, 'packageName')
    event_data.prediction_rank = self._GetRowValue(
        query_hash, row, 'predictionRank')
    event_data.prediction_source_identifier = self._GetRowValue(
        query_hash, row, 'predictionSourceId')
    event_data.prediction_ui_surface_identifier = self._GetRowValue(
        query_hash, row, 'predictionUiSurfaceId')
    event_data.start_time = self._GetTimeRowValue(
        query_hash, row, 'timestampMillis')


    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(AndroidAppLaunchPlugin)



"""Parsers for MacOS fseventsd files.

Also see:
  https://github.com/libyal/dtformats/blob/main/documentation/MacOS%20File%20System%20Events%20Disk%20Log%20Stream%20format.asciidoc
"""

import os

from dfvfs.resolver import resolver as path_spec_resolver

from plaso.containers import events
from plaso.lib import dtfabric_helper
from plaso.lib import errors
from plaso.lib import specification
from plaso.parsers import interface
from plaso.parsers import manager


class FseventsdEventData(events.EventData):
  """MacOS file system event (fseventsd) event data.

  Attributes:
    event_identifier (int): the record event identifier.
    file_entry_modification_time (dfdatetime.DateTimeValues): file entry
        last modification date and time.
    flags (int): flags stored in the record.
    node_identifier (int): file system node identifier related to the file
        system event.
    path (str): path recorded in the fseventsd record.
  """

  DATA_TYPE = 'macos:fseventsd:record'

  def __init__(self):
    """Initializes an Fseventsd event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.event_identifier = None
    self.file_entry_modification_time = None
    self.flags = None
    self.node_identifier = None
    self.path = None


class FseventsdParser(
    interface.FileObjectParser, dtfabric_helper.DtFabricHelper):
  """Parser for fseventsd files.

  This parser supports both version 1 and version 2 fseventsd files.
  """

  NAME = 'fseventsd'
  DATA_FORMAT = 'MacOS File System Events Disk Log Stream (fseventsd) file'

  # The version 1 format was used in Mac OS X 10.5 (Leopard) through macOS 10.12
  # (Sierra).
  _DLS_V1_SIGNATURE = b'1SLD'

  # The version 2 format was introduced in MacOS High Sierra (10.13).
  _DLS_V2_SIGNATURE = b'2SLD'

  _DEFINITION_FILE = os.path.join(
      os.path.dirname(__file__), 'fseventsd.yaml')

  @classmethod
  def GetFormatSpecification(cls):
    """Retrieves the format specification.

    Returns:
      FormatSpecification: format specification.
    """
    format_specification = specification.FormatSpecification(cls.NAME)
    format_specification.AddNewSignature(cls._DLS_V1_SIGNATURE, offset=0)
    format_specification.AddNewSignature(cls._DLS_V2_SIGNATURE, offset=0)
    return format_specification

  def _ParseDLSPageHeader(self, file_object, page_offset):
    """Parses a DLS page header from a file-like object.

    Args:
      file_object (file): file-like object to read the header from.
      page_offset (int): offset of the start of the page header, relative
          to the start of the file.

    Returns:
      tuple: containing:

        dls_page_header: parsed record structure.
        int: header size.

    Raises:
      ParseError: when the header cannot be parsed.
    """
    page_header_map = self._GetDataTypeMap('dls_page_header')

    try:
      page_header, page_size = self._ReadStructureFromFileObject(
          file_object, page_offset, page_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.ParseError(
          f'Unable to parse page header at offset: 0x{page_offset:08x} '
          f'with error: {exception!s}')

    return page_header, page_size

  def _BuildEventData(self, record):
    """Builds an FseventsdData object from a parsed structure.

    Args:
      record (dls_record_v1|dls_record_v2): parsed record structure.

    Returns:
      FseventsdEventData: event data attribute container.
    """
    event_data = FseventsdEventData()
    event_data.path = record.path
    event_data.flags = record.event_flags
    event_data.event_identifier = record.event_identifier
    # Node identifier is only set in DLS V2 records.
    event_data.node_identifier = getattr(record, 'node_identifier', None)

    return event_data

  def _GetParentModificationTime(self, gzip_file_entry):
    """Retrieves the modification time of the file entry's parent file.

    Note that this retrieves the time from the file entry of the parent of the
    gzip file entry's path spec, which is different from trying to retrieve it
    from the gzip file entry's parent file entry.

    It would be preferable to retrieve the modification time from the metadata
    in the gzip file itself, but it appears to not be set when the file is
    written by fseventsd.

    Args:
      gzip_file_entry (dfvfs.FileEntry): file entry of the gzip file containing
          the fseventsd data.

    Returns:
      dfdatetime.DateTimeValues: parent modification time, or None if not
          available.
    """
    parent_file_entry = path_spec_resolver.Resolver.OpenFileEntry(
        gzip_file_entry.path_spec.parent)
    if not parent_file_entry:
      return None

    return parent_file_entry.modification_time

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses an fseventsd file.

    Args:
      parser_mediator (ParserMediator): parser mediator.
      file_object (dfvfs.FileIO): a file-like object.

    Raises:
      WrongParser: when the header cannot be parsed.
    """
    page_header_map = self._GetDataTypeMap('dls_page_header')

    try:
      page_header, file_offset = self._ReadStructureFromFileObject(
          file_object, 0, page_header_map)
    except (ValueError, errors.ParseError) as exception:
      raise errors.WrongParser(
          f'Unable to parse page header with error: {exception!s}')

    current_page_end = page_header.page_size

    file_entry = parser_mediator.GetFileEntry()
    date_time = self._GetParentModificationTime(file_entry)

    file_size = file_object.get_size()
    while file_offset < file_size:
      if file_offset >= current_page_end:
        try:
          page_header, header_size = self._ParseDLSPageHeader(
              file_object, file_offset)
        except errors.ParseError as exception:
          parser_mediator.ProduceExtractionWarning(
              f'Unable to parse page header with error: {exception!s}')
          break

        current_page_end += page_header.page_size
        file_offset += header_size
        continue

      if page_header.signature == self._DLS_V1_SIGNATURE:
        record_map = self._GetDataTypeMap('dls_record_v1')
      else:
        record_map = self._GetDataTypeMap('dls_record_v2')

      try:
        record, record_length = self._ReadStructureFromFileObject(
            file_object, file_offset, record_map)
        file_offset += record_length
      except (ValueError, errors.ParseError) as exception:
        parser_mediator.ProduceExtractionWarning(
            f'Unable to parse page record with error: {exception!s}')
        break

      event_data = self._BuildEventData(record)
      event_data.file_entry_modification_time = date_time

      parser_mediator.ProduceEventData(event_data)


manager.ParsersManager.RegisterParser(FseventsdParser)



"""SQLite parser plugin for "Files by Google" database (Web Data) files.

The Files by Google app database file is typically stored in:
  /data/user/0/com.google.android.apps.nbu.files/databases/files_master_database

This SQLite database is used by the Files by Google app to manage and organize
files on Android devices. The database contains information on file metadata,
including file names, paths, sizes, and types, as well as records of storage
locations.
"""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class FilesByGoogleEventData(events.EventData):
  """Files by Google event data.

  Attributes:
    file_name (str): name of a file.
    file_size (int): size of a file.
    is_hidden (int): mark of a hidden file.
    media_type (str): file's media type.
    mime_type (str): format of a file.
    modification_time (dfdatetime.DateTimeValues): modification date and time
        of a file.
    parent_folder (str): file's parent folder.
    query (str): SQL query that was used to obtain the event data.
    root_path (str): memory location.
    root_relative_path (str): root path of a file.
    title (str): name of a file.
    uri (str): character sequence that identifies a resource.
  """

  DATA_TYPE = 'android:files_by_google:record'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.file_name = None
    self.file_size = None
    self.is_hidden = None
    self.media_type = None
    self.mime_type = None
    self.modification_time = None
    self.parent_folder = None
    self.query = None
    self.root_path = None
    self.root_relative_path = None
    self.title = None
    self.uri = None


class FilesByGooglePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Google "Files by Google" database files."""

  NAME = 'files_by_google'
  DATA_FORMAT = 'Files by Google SQLite database (files_master_database) file'

  REQUIRED_STRUCTURE = {
      'files_master_table': frozenset([
          'file_date_modified_ms', 'root_path', 'root_relative_file_path',
          'file_name', 'size', 'mime_type', 'media_type', 'uri', 'is_hidden',
          'title', 'parent_folder_name'])}

  QUERIES = [(
      ('SELECT files_master_table.file_date_modified_ms,'
       'files_master_table.root_path,'
       'files_master_table.root_relative_file_path,'
       'files_master_table.file_name, files_master_table.size,'
       'files_master_table.mime_type, files_master_table.media_type,'
       'files_master_table.uri,files_master_table.is_hidden,'
       'files_master_table.title, files_master_table.parent_folder_name '
       'FROM files_master_table'), '_ParseFilesByGoogleRow')]

  SCHEMAS = [{
      'android_metadata':(
          'CREATE TABLE android_metadata (locale TEXT)'),
      'file_search_fts':(
           'CREATE VIRTUAL TABLE file_search_fts USING '
           'fts4(content="files_master_table", root_relative_file_path, '
           'title, artist, album)'),
      'file_search_fts_docsize':(
           'CREATE TABLE file_search_fts_docsize(docid INTEGER PRIMARY KEY, '
           'size BLOB)'),
      'file_search_fts_segdir':(
           'CREATE TABLE file_search_fts_segdir(level INTEGER, idx INTEGER, '
           'start_block INTEGER, leaves_end_block INTEGER, '
           'end_block INTEGER, root BLOB, '
           'PRIMARY KEY(level, idx))'),
      'file_search_fts_segments':(
           'CREATE TABLE file_search_fts_segments(blockid INTEGER PRIMARY KEY, '
           'block BLOB)'),
      'file_search_fts_stat':(
           'CREATE TABLE file_search_fts_stat(id INTEGER PRIMARY KEY, '
           'value BLOB)'),
      'files_master_table':(
           'CREATE TABLE files_master_table(id INTEGER PRIMARY KEY '
           'AUTOINCREMENT, media_store_id INTEGER, root_path TEXT NOT NULL '
           'DEFAULT \'\', root_relative_file_path TEXT NOT NULL DEFAULT \'\', '
           'file_name TEXT NOT NULL, size INTEGER NOT NULL, '
           'file_date_modified_ms INTEGER NOT NULL, '
           'storage_location INTEGER NOT NULL, mime_type TEXT, '
           'media_type INTEGER, uri TEXT NOT NULL, '
           'is_hidden INTEGER NOT NULL, title TEXT, '
           'artist TEXT, album TEXT, '
           'parent_folder_name TEXT COLLATE NOCASE, UNIQUE (media_store_id), '
           'UNIQUE (uri))'),
      'media_store_scan_state_table':(
           'CREATE TABLE media_store_scan_state_table(id INTEGER NOT NULL '
           'DEFAULT 0 UNIQUE, largest_scanned_date_modified_ms INTEGER NOT '
           'NULL, largest_scanned_media_store_id INTEGER NOT NULL)'),
      'post_file_scan_tasks_state_table':(
           'CREATE TABLE post_file_scan_tasks_state_table(task_type INTEGER '
           'NOT NULL UNIQUE, largest_processed_file_id INTEGER NOT NULL, '
           'last_processed_timestamp_ms INTEGER NOT NULL)'),
      'scan_log_table':(
           'CREATE TABLE scan_log_table(id INTEGER PRIMARY KEY AUTOINCREMENT, '
           'start_timestamp INTEGER NOT NULL DEFAULT 0, '
           'end_timestamp INTEGER NOT NULL DEFAULT 0, '
           'updated_items_count INTEGER NOT NULL DEFAULT 0, scan_type INTEGER '
           'NOT NULL, scan_result INTEGER NOT NULL DEFAULT 0)'),
      'file_consumption_data_table':(
           'CREATE TABLE file_consumption_data_table(id INTEGER PRIMARY KEY '
           'AUTOINCREMENT, main_file_table_id INTEGER NOT NULL, '
           'play_state INTEGER, duration INTEGER, play_position INTEGER, '
           'fully_played_count INTEGER NOT NULL DEFAULT 0, '
           'FOREIGN KEY(main_file_table_id) '
           'REFERENCES files_master_table(id) ON DELETE CASCADE)'),
      'file_open_info_table':(
           'CREATE TABLE file_open_info_table(file_id INTEGER NOT NULL, '
           'file_open_time_stamp_ms INTEGER NOT NULL,  UNIQUE(file_id, '
           'file_open_time_stamp_ms), FOREIGN KEY(file_id) REFERENCES '
           'files_master_table(id) ON DELETE CASCADE)'),
      'files_classification_table':(
           'CREATE TABLE files_classification_table(file_id INTEGER NOT NULL, '
           'classification INTEGER NOT NULL DEFAULT 0, FOREIGN KEY(file_id) '
           'REFERENCES files_master_table(id) ON DELETE CASCADE)'),
      'files_metadata_table':(
           'CREATE TABLE files_metadata_table(file_id INTEGER UNIQUE NOT NULL, '
           'title TEXT, artist TEXT, '
           'album TEXT, is_drm_protected INTEGER NOT NULL DEFAULT 0, '
           'latitude REAL, longitude REAL, '
           'width_px INTEGER, height_px INTEGER, '
           'duration_px INTEGER, track_number INTEGER, '
           'file_hash TEXT, taken_date_ms INTEGER, '
           'device_description TEXT, aperture DOUBLE, '
           'focal_length DOUBLE, exposure_time DOUBLE, '
           'iso_speed INTEGER, FOREIGN KEY(file_id) REFERENCES '
           'files_master_table(id) ON DELETE CASCADE)'),
      'nima_score_table':(
           'CREATE TABLE nima_score_table(file_id INTEGER NOT NULL, '
           'vg_score FLOAT NOT NULL, ava_score FLOAT NOT NULL, '
           'UNIQUE (file_id), FOREIGN KEY(file_id) REFERENCES '
           'files_master_table(id) ON DELETE CASCADE)'),
      'session_scan_log_table':(
           'CREATE TABLE session_scan_log_table(scan_id INTEGER NOT NULL, '
           'item_id INTEGER NOT NULL, FOREIGN KEY(scan_id) REFERENCES '
           'scan_log_table(id), FOREIGN KEY(item_id) REFERENCES '
           'files_master_table(id))')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.PosixTimeInMilliseconds: date and time value or None if
      not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_posix_time.PosixTimeInMilliseconds(timestamp=timestamp)

  def _ParseFilesByGoogleRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses an files by google entry row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = FilesByGoogleEventData()
    event_data.file_name = self._GetRowValue(query_hash, row, 'file_name')
    event_data.file_size = self._GetRowValue(query_hash, row, 'size')
    event_data.is_hidden = self._GetRowValue(query_hash, row, 'is_hidden')
    event_data.media_type = self._GetRowValue(query_hash, row, 'media_type')
    event_data.mime_type = self._GetRowValue(query_hash, row, 'mime_type')
    event_data.modification_time = self._GetDateTimeRowValue(
        query_hash, row, 'file_date_modified_ms')
    event_data.parent_folder = self._GetRowValue(
        query_hash, row, 'parent_folder_name')
    event_data.query = query
    event_data.root_path = self._GetRowValue(query_hash, row, 'root_path')
    event_data.root_relative_path = self._GetRowValue(
        query_hash, row, 'root_relative_file_path')
    event_data.title = self._GetRowValue(query_hash, row, 'title')
    event_data.uri = self._GetRowValue(query_hash, row, 'uri')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(FilesByGooglePlugin)



"""Shared functionality for delimiter separated values output modules."""

import collections

from plaso.output import formatting_helper
from plaso.output import text_file


class DSVEventFormattingHelper(formatting_helper.EventFormattingHelper):
  """Delimiter separated values output module event formatting helper.

  Attributes:
    field_delimiter (str): field delimiter.
  """

  def __init__(self, field_formatting_helper, field_names, field_delimiter=','):
    """Initializes a delimiter separated values event formatting helper.

    Args:
      field_formatting_helper (FieldFormattingHelper): field formatting helper.
      field_names (list[str]): names of the fields to output.
      field_delimiter (Optional[str]): field delimiter.
    """
    super().__init__()
    self._custom_fields = {}
    self._field_names = field_names
    self._field_formatting_helper = field_formatting_helper

    self.field_delimiter = field_delimiter

  def _SanitizeField(self, field):
    """Sanitizes a field for output.

    This method replaces any field delimiters with a space.

    Args:
      field (str): value of the field to sanitize.

    Returns:
      str: sanitized value of the field.
    """
    if self.field_delimiter and isinstance(field, str):
      return field.replace(self.field_delimiter, ' ')
    return field

  def GetFieldValues(
      self, output_mediator, event, event_data, event_data_stream, event_tag):
    """Retrieves the output field values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      dict[str, str]: output field values per name.
    """
    field_values = collections.OrderedDict()
    for field_name in self._field_names:
      field_value = self._field_formatting_helper.GetFormattedField(
          output_mediator, field_name, event, event_data, event_data_stream,
          event_tag)

      if field_value is None and field_name in self._custom_fields:
        field_value = self._custom_fields.get(field_name)

      if field_value is None:
        field_value = '-'

      field_value = self._SanitizeField(field_value)
      field_values[field_name] = field_value

    return field_values

  def GetFormattedFieldNames(self):
    """Retrieves a string representation of the field names.

    Returns:
      str: string representation of the field names.
    """
    return self.field_delimiter.join(self._field_names)

  def SetAdditionalFields(self, field_names):
    """Sets the names of additional fields to output.

    Args:
      field_names (list[str]): names of additional fields to output.
    """
    self._field_names.extend(field_names)

  def SetCustomFields(self, field_names_and_values):
    """Sets the names and values of custom fields to output.

    Args:
      field_names_and_values (list[tuple[str, str]]): names and values of
          custom fields to output.
    """
    self._custom_fields = dict(field_names_and_values)
    self._field_names.extend(self._custom_fields.keys())

  def SetFieldDelimiter(self, field_delimiter):
    """Sets the field delimiter.

    Args:
      field_delimiter (str): field delimiter.
    """
    self.field_delimiter = field_delimiter

  def SetFields(self, field_names):
    """Sets the names of the fields to output.

    Args:
      field_names (list[str]): names of the fields to output.
    """
    self._field_names = field_names


class DSVOutputModule(text_file.SortedTextFileOutputModule):
  """Shared functionality for delimiter separated values output modules."""

  def __init__(
      self, field_formatting_helper, names, delimiter=',', header=None):
    """Initializes a delimiter separated values output module.

    Args:
      field_formatting_helper (FieldFormattingHelper): field formatting helper.
      names (list[str]): names of the fields to output.
      delimiter (Optional[str]): field delimiter.
      header (Optional[str]): header, where None will have WriteHeader
          generate a header from the field names.
    """
    event_formatting_helper = DSVEventFormattingHelper(
        field_formatting_helper, names, field_delimiter=delimiter)
    super().__init__(event_formatting_helper)
    self._header = header

  def _GetString(self, output_mediator, field_values):
    """Retrieves an output string.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.

    Returns:
      str: output string.
    """
    output_text = self._event_formatting_helper.field_delimiter.join(
        field_values.values())
    return ''.join([output_text, '\n'])

  def SetAdditionalFields(self, field_names):
    """Sets the names of additional fields to output.

    Args:
      field_names (list[str]): names of additional fields to output.
    """
    self._event_formatting_helper.SetAdditionalFields(field_names)

  def SetCustomFields(self, field_names_and_values):
    """Sets the names and values of custom fields to output.

    Args:
      field_names_and_values (list[tuple[str, str]]): names and values of
          custom fields to output.
    """
    self._event_formatting_helper.SetCustomFields(field_names_and_values)

  def SetFieldDelimiter(self, field_delimiter):
    """Sets the field delimiter.

    Args:
      field_delimiter (str): field delimiter.
    """
    self._event_formatting_helper.SetFieldDelimiter(field_delimiter)

  def SetFields(self, field_names):
    """Sets the names of the fields to output.

    Args:
      field_names (list[str]): names of the fields to output.
    """
    self._event_formatting_helper.SetFields(field_names)

  def WriteHeader(self, output_mediator):
    """Writes the header to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
    """
    if self._header:
      output_text = self._header
    else:
      output_text = self._event_formatting_helper.GetFormattedFieldNames()

    self.WriteLine(output_text)



"""Output module field formatting helper."""

import abc
import datetime
import math
import pytz

from dfdatetime import posix_time as dfdatetime_posix_time
from dfvfs.lib import definitions as dfvfs_definitions

from plaso.containers import events
from plaso.formatters import default
from plaso.output import logger


class EventFormattingHelper:
  """Output module event formatting helper."""

  @abc.abstractmethod
  def GetFieldValues(
      self, output_mediator, event, event_data, event_data_stream, event_tag):
    """Retrieves the output field values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      list[str]: output field values.
    """


class FieldFormattingHelper:
  """Output module field formatting helper."""

  _DEFAULT_MESSAGE_FORMATTER = default.DefaultEventFormatter()

  # Maps the name of a field to callback function that formats the field value.
  _FIELD_FORMAT_CALLBACKS = {}

  def __init__(self):
    """Initializes a field formatting helper."""
    event_data_stream = events.EventDataStream()

    super().__init__()
    self._callback_functions = {}
    self._event_data_stream_field_names = event_data_stream.GetAttributeNames()
    self._event_tag_field_names = []

    for field_name, callback_name in self._FIELD_FORMAT_CALLBACKS.items():
      if callback_name == '_FormatTag':
        self._event_tag_field_names.append(field_name)
      else:
        self._callback_functions[field_name] = getattr(
            self, callback_name, None)

  # The field format callback methods require specific arguments hence
  # the check for unused arguments is disabled here.
  # pylint: disable=unused-argument

  def _FormatDateTime(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a date and time field in ISO 8601 format.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: date and time field with time zone offset, semantic time.
    """
    if output_mediator.dynamic_time and event.date_time:
      iso8601_string = getattr(event.date_time, 'string', None)
      if iso8601_string:
        return iso8601_string

      date_time = event.date_time
      if event.date_time.is_local_time:
        # TODO: replace this by a more generic solution.
        if event.date_time.precision == '1s':
          timestamp, _ = divmod(event.timestamp, 1000000)
          date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
        else:
          date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
              timestamp=event.timestamp)

      iso8601_string = date_time.CopyToDateTimeStringISO8601()
      if not iso8601_string:
        return 'Invalid'

      if iso8601_string[-1] == 'Z':
        iso8601_string = f'{iso8601_string[:-1]:s}+00:00'

      if output_mediator.time_zone != pytz.UTC or date_time.time_zone_offset:
        # For output in a specific time zone overwrite the date, time in
        # seconds and time zone offset in the UTC ISO8601 string.
        year, month, day_of_month, hours, minutes, seconds = (
            date_time.GetDateWithTimeOfDay())

        try:
          datetime_object = datetime.datetime(
              year, month, day_of_month, hours, minutes, seconds,
              tzinfo=pytz.UTC)

          datetime_object = datetime_object.astimezone(
              output_mediator.time_zone)

          isoformat_string = datetime_object.isoformat()
          iso8601_string = ''.join([
              isoformat_string[:19], iso8601_string[19:-6],
              isoformat_string[-6:]])
        except (OSError, OverflowError, TypeError, ValueError):
          return 'Invalid'

    else:
      if not event.date_time or event.date_time.is_local_time:
        timestamp = event.timestamp
      else:
        timestamp, fraction_of_second = (
            event.date_time.CopyToPosixTimestampWithFractionOfSecond())

        timestamp = (timestamp or 0) * 1000000

        if fraction_of_second:
          while fraction_of_second < 1000000:
            fraction_of_second *= 10
          while fraction_of_second >= 1000000:
            fraction_of_second /= 10

          if timestamp < 0:
            timestamp -= math.ceil(fraction_of_second)
          else:
            timestamp += math.ceil(fraction_of_second)

      # For now check if event.timestamp is set, to mimic existing behavior of
      # using 0000-00-00T00:00:00.000000+00:00 for 0 timestamp values
      if not timestamp:
        return '0000-00-00T00:00:00.000000+00:00'

      try:
        # Note that setting tzinfo to None can have unexpected results on
        # certain platforms.
        datetime_object = (
            datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC) +
            datetime.timedelta(microseconds=timestamp))

        datetime_object = datetime_object.astimezone(output_mediator.time_zone)

        iso8601_string = datetime_object.isoformat()
        iso8601_string = (
            f'{iso8601_string[:19]:s}.{datetime_object.microsecond:06d}'
            f'{iso8601_string[-6:]:s}')

      except (OSError, OverflowError, TypeError, ValueError) as exception:
        iso8601_string = '0000-00-00T00:00:00.000000+00:00'
        self._ReportEventError(event, event_data, (
            f'unable to copy timestamp: {event.timestamp!s} to a human '
            f'readable date and time with error: {exception!s}. Defaulting '
            f'to: "{iso8601_string:s}"'))

    return iso8601_string

  def _FormatDisplayName(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats the display name.

    The display_name field can be set as an attribute to event_data otherwise
    it is derived from the path specification.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: display name field.
    """
    display_name = getattr(event_data, 'display_name', None)
    if not display_name:
      path_spec = getattr(event_data_stream, 'path_spec', None)
      if path_spec:
        display_name = output_mediator.GetDisplayNameForPathSpec(path_spec)
      else:
        display_name = '-'

    return display_name

  def _FormatFilename(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats the filename.

    The filename field can be set as an attribute to event_data otherwise
    it is derived from the path specification.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: filename field.
    """
    filename = getattr(event_data, 'filename', None)
    if not filename:
      path_spec = getattr(event_data_stream, 'path_spec', None)
      if path_spec:
        filename = output_mediator.GetRelativePathForPathSpec(path_spec)
      else:
        filename = '-'

    return filename

  def _FormatHostname(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a hostname field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: hostname field.
    """
    return output_mediator.GetHostname(event_data)

  def _FormatInode(self, output_mediator, event, event_data, event_data_stream):
    """Formats an inode field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: inode field.
    """
    inode = getattr(event_data, 'inode', None)

    # Note that inode can contain 0.
    if inode is None:
      path_specification = getattr(event_data_stream, 'path_spec', None)
      if path_specification:
        if path_specification.type_indicator in (
            dfvfs_definitions.TYPE_INDICATOR_APFS,
            dfvfs_definitions.TYPE_INDICATOR_HFS):
          inode = getattr(path_specification, 'identifier', None)

        elif path_specification.type_indicator == (
            dfvfs_definitions.TYPE_INDICATOR_NTFS):
          inode = getattr(path_specification, 'mft_entry', None)

        elif path_specification.type_indicator in (
            dfvfs_definitions.TYPE_INDICATOR_EXT,
            dfvfs_definitions.TYPE_INDICATOR_TSK):
          # Note that inode can contain a TSK metadata address.
          inode = getattr(path_specification, 'inode', None)

    if inode is None:
      inode = '-'

    elif isinstance(inode, int):
      inode = f'{inode:d}'

    return inode

  def _FormatMACB(self, output_mediator, event, event_data, event_data_stream):
    """Formats a legacy MACB representation field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: MACB field.
    """
    return output_mediator.GetMACBRepresentation(event, event_data)

  def _FormatMessage(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a message field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: message field.
    """
    message_formatter = output_mediator.GetMessageFormatter(
        event_data.data_type)
    if not message_formatter:
      logger.warning(
          f'Using default message formatter for data type: '
          f'{event_data.data_type:s}')
      message_formatter = self._DEFAULT_MESSAGE_FORMATTER

    event_values = event_data.CopyToDict()
    message_formatter.FormatEventValues(output_mediator, event_values)

    return message_formatter.GetMessage(event_values)

  def _FormatMessageShort(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a short message field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: short message field.
    """
    message_formatter = output_mediator.GetMessageFormatter(
        event_data.data_type)
    if not message_formatter:
      logger.warning(
          f'Using default message formatter for data type: '
          f'{event_data.data_type:s}')
      message_formatter = self._DEFAULT_MESSAGE_FORMATTER

    event_values = event_data.CopyToDict()
    message_formatter.FormatEventValues(output_mediator, event_values)

    return message_formatter.GetMessageShort(event_values)

  def _FormatParser(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a parser field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: parser field.
    """
    return getattr(event_data, '_parser_chain', None) or '-'

  def _FormatSource(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a source field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: source field.
    """
    data_type = getattr(event_data, 'data_type', None) or '-'
    _, source = output_mediator.GetSourceMapping(data_type)
    return source or 'N/A'

  def _FormatSourceShort(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a short source field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: short source field.
    """
    data_type = getattr(event_data, 'data_type', None) or '-'
    source_short, _ = output_mediator.GetSourceMapping(data_type)
    return source_short or 'N/A'

  def _FormatTag(self, output_mediator, event_tag):
    """Formats an event tag field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event_tag (EventTag): event tag or None if not set.

    Returns:
      str: event tag labels or "-" if event tag is not set.
    """
    if not event_tag:
      return '-'

    return ' '.join(event_tag.labels)

  def _FormatTime(self, output_mediator, event, event_data, event_data_stream):
    """Formats a time field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: time in seconds formatted as "HH:MM:SS" or "--:--:--" on error.
    """
    # For now check if event.timestamp is set, to mimic existing behavior of
    # using --:--:-- for 0 timestamp values.
    if not event.timestamp:
      return '--:--:--'

    date_time = event.date_time
    if not date_time or date_time.is_local_time:
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=event.timestamp)

    year, month, day_of_month, hours, minutes, seconds = (
        date_time.GetDateWithTimeOfDay())

    if output_mediator.time_zone != pytz.UTC:
      try:
        datetime_object = datetime.datetime(
            year, month, day_of_month, hours, minutes, seconds,
            tzinfo=pytz.UTC)

        datetime_object = datetime_object.astimezone(output_mediator.time_zone)

        hours, minutes, seconds = (
            datetime_object.hour, datetime_object.minute,
            datetime_object.second)

      except (OSError, OverflowError, TypeError, ValueError):
        hours, minutes, seconds = (None, None, None)

    if None in (hours, minutes, seconds):
      self._ReportEventError(event, event_data, (
          f'unable to copy timestamp: {event.timestamp!s} to a human readable '
          f'time. Defaulting to: "--:--:--"'))
      return '--:--:--'

    return f'{hours:02d}:{minutes:02d}:{seconds:02d}'

  def _FormatTimeZone(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a time zone field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: time zone field.
    """
    if not event.timestamp:
      return '-'

    date_time = event.date_time
    if not date_time or date_time.is_local_time:
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=event.timestamp)

    if output_mediator.time_zone == pytz.UTC:
      return 'UTC'

    year, month, day_of_month, hours, minutes, seconds = (
        date_time.GetDateWithTimeOfDay())

    try:
      # For tzname to work the datetime object must be naive (without
      # a time zone).
      datetime_object = datetime.datetime(
          year, month, day_of_month, hours, minutes, seconds)
      return output_mediator.time_zone.tzname(datetime_object)

    except (OverflowError, TypeError, ValueError):
      self._ReportEventError(event, event_data, (
          f'unable to copy timestamp: {event.timestamp!s} to a human readable '
          f'time zone. Defaulting to: "-"'))
      return '-'

  def _FormatUsername(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats an username field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: username field.
    """
    return output_mediator.GetUsername(event_data)

  def _FormatValues(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a values.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: values field.
    """
    values = getattr(event_data, 'values', None)
    if isinstance(values, list) and event_data.data_type in (
        'windows:registry:key_value', 'windows:registry:service'):
      values = ' '.join(
          f'{name or "(default)"}: [{data_type:s}] {data or "(empty)"}'
          for name, data_type, data in sorted(values))

    return values

  # pylint: enable=unused-argument

  def _ReportEventError(self, event, event_data, error_message):
    """Reports an event related error.

    Args:
      event (EventObject): event.
      event_data (EventData): event data.
      error_message (str): error message.
    """
    event_identifier = event.GetIdentifier()
    event_identifier_string = event_identifier.CopyToString()
    display_name = getattr(event_data, 'display_name', None) or 'N/A'
    parser_chain = getattr(event_data, '_parser_chain', None) or 'N/A'

    logger.error((
        f'Event: {event_identifier_string!s} description: '
        f'{event.timestamp_desc:s} data type: {event_data.data_type:s} '
        f'display name: {display_name:s} parser chain: {parser_chain:s} with '
        f'error: {error_message:s}'))

  def GetFormattedField(
      self, output_mediator, field_name, event, event_data, event_data_stream,
      event_tag):
    """Formats the specified field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_name (str): name of the field.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      event_tag (EventTag): event tag.

    Returns:
      str: value of the field or None if not available.
    """
    if field_name in self._event_tag_field_names:
      return self._FormatTag(output_mediator, event_tag)

    callback_function = self._callback_functions.get(field_name)
    if callback_function:
      output_value = callback_function(
          output_mediator, event, event_data, event_data_stream)
    elif field_name in self._event_data_stream_field_names:
      output_value = getattr(event_data_stream, field_name, None)
    else:
      output_value = getattr(event_data, field_name, None)

    if output_value is not None and not isinstance(output_value, str):
      output_value = f'{output_value!s}'

    return output_value



"""Plist parser plugin for Apple iOS Car Play application plist files.

The plist contains history of opened applications in the Car Play application.
"""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.lib import definitions
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class IOSCarPlayHistoryEventData(events.EventData):
  """Apple iOS Car Play application history event data.

  Attributes:
    application_identifier (str): application identifier.
    last_run_time (dfdatetime.DateTimeValues): application last run date and
        time.
  """

  DATA_TYPE = 'ios:carplay:history:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.application_identifier = None
    self.last_run_time = None


class IOSCarPlayPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Apple iOS Car Play application plist files."""

  NAME = 'ios_carplay'
  DATA_FORMAT = 'Apple iOS Car Play application plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('com.apple.CarPlayApp.plist')])

  PLIST_KEYS = frozenset(['CARRecentAppHistory'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extract Car Play application history entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    plist_key = match.get('CARRecentAppHistory', {})
    for application_identifier, datetime_value in plist_key.items():
      event_data = IOSCarPlayHistoryEventData()
      event_data.application_identifier = application_identifier

      if datetime_value:
        timestamp = int(datetime_value * definitions.NANOSECONDS_PER_SECOND)
        event_data.last_run_time = dfdatetime_posix_time.PosixTimeInNanoseconds(
            timestamp=timestamp)

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(IOSCarPlayPlistPlugin)



"""MacOS Darwin helpers."""


class DarwinSystemErrorHelper:
  """Darwin system error helper.

  Also see:
    https://github.com/apple/darwin-xnu/blob/main/bsd/sys/errno.h
  """

  _ERROR_CODES = {
      -1: 'Restart syscall',
      -2: 'Don\'t modify regs, just return',
      -5: 'Restart lookup under heavy vnode pressure/recycling',
      -6: 'Red drive open',
      -7: 'Keep looking',
      -8: 'Data less',
      0: 'Success',
      1: 'Operation not permitted',
      2: 'No such file or directory',
      3: 'No such process',
      4: 'Interrupted system call',
      5: 'Input/Output error',
      6: 'Device not configured',
      7: 'Argument list too long',
      8: 'Exec format error',
      9: 'Bad file descriptor',
      10: 'No child processes',
      11: 'Resource deadlock avoided',
      12: 'Cannot allocate memory',
      13: 'Permission denied',
      14: 'Bad address',
      15: 'Block device required',
      16: 'Device / Resource busy',
      17: 'File exists',
      18: 'Cross-device link',
      19: 'Operation not supported by device',
      20: 'Not a directory',
      21: 'Is a directory',
      22: 'Invalid argument',
      23: 'Too many open files in system',
      24: 'Too many open files',
      25: 'Inappropriate ioctl for devices',
      26: 'Text file busy',
      27: 'File too large',
      28: 'No space left on device',
      29: 'Illegal seek',
      30: 'Read-only filesystem',
      31: 'Too many links',
      32: 'Broken pipe',
      33: 'Numerical argument out of domain',
      34: 'Result too large',
      35: 'Resource temporarily unavailable, operation would block',
      36: 'Operation now in progress',
      37: 'Operation already in progress',
      38: 'Socket operation on non-socket',
      39: 'Destination address required',
      40: 'Message too long',
      41: 'Protocol wrong type for socket',
      42: 'Protocol not available',
      43: 'Protocol not supported',
      44: 'Socket type not supported',
      45: 'Operation not supported, Operation not supported on socket',
      46: 'Protocol family not supported',
      47: 'Address family not supported by protocol family',
      48: 'Address already in use',
      49: 'Can\'t assign requested address',
      50: 'Network is down',
      51: 'Network is unreachable',
      52: 'Network dropped connection on reset',
      53: 'Software caused connection abort',
      54: 'Connection reset by peer',
      55: 'No buffer space available',
      56: 'Socket is already connected',
      57: 'Socket is not connected',
      58: 'Can\'t send after socket shutdown',
      59: 'Too many references: can\'t splice',
      60: 'Operation timed out',
      61: 'Connection refused',
      62: 'Too many levels of symbolic links',
      63: 'File name too long',
      64: 'Host is down',
      65: 'No route to host',
      66: 'Directory not empty',
      67: 'Too many processes',
      68: 'Too many users',
      69: 'Disc quota exceeded',
      70: 'Stale NFS file handle',
      71: 'Too many levels of remote in path',
      72: 'RPC struct is bad',
      73: 'RPC version wrong',
      74: 'RPC prog. not avail',
      75: 'Program version wrong',
      76: 'Bad procedure for program',
      77: 'No locks available',
      78: 'Function not implemented',
      79: 'Inappropriate file type or format',
      80: 'Authentication error',
      81: 'Need authenticator',
      82: 'Device power is off',
      83: 'Device error, e.g. paper out',
      84: 'Value too large to be stored in data type',
      85: 'Bad executable',
      86: 'Bad CPU type in executable',
      87: 'Shared library version mismatch',
      88: 'Malformed Macho file',
      89: 'Operation canceled',
      90: 'Identifier removed',
      91: 'No message of desired type',
      92: 'Illegal byte sequence',
      93: 'Attribute not found',
      94: 'Bad message',
      95: 'Reserved',
      96: 'No message available on STREAM',
      97: 'Reserved',
      98: 'No STREAM resources',
      99: 'Not a STREAM',
      100: 'Protocol error',
      101: 'STREAM ioctl timeout',
      102: 'Operation not supported on socket',
      103: 'No such policy registered',
      104: 'State not recoverable',
      105: 'Previous owner died',
      106: 'Interface output queue is full, Must be equal largest errno'}

  @classmethod
  def GetDescription(cls, error_code):
    """Retrieves the description of a specific error code.

    Args:
      error_code (int): error code.

    Returns:
      str: description of the error.
    """
    return cls._ERROR_CODES.get(error_code, f'UNKNOWN: {error_code:d}')
